#    Copyright 2025 Genesis Corporation.
#    Copyright 2026 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg

CONF = cfg.CONF


DOMAIN_IAM = "iam"


def register_iam_cli_opts(conf):

    conf = cfg.CONF

    iam_cli_opts = [
        cfg.StrOpt(
            "iam_endpoint",
            default=(
                "http://core.local.genesis-core.tech:11010/v1/iam/clients/"
                "00000000-0000-0000-0000-000000000000"
            ),
            help="IAM endpoint used by services",
        ),
        cfg.StrOpt(
            "audience",
            default="GenesisCoreClientId",
            help="The correct audience of the JWT token for this service",
        ),
        cfg.StrOpt(
            "hs256_jwks_decryption_key",
            default="AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
            secret=True,
            help="HS256 JWKS decryption key (A256GCM key, base64 or utf-8)",
        ),
    ]

    conf.register_cli_opts(iam_cli_opts, DOMAIN_IAM)



# Copyright 2025 Genesis Corporation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import base64
import binascii
import logging
import os
import typing as tp

from cryptography.hazmat.primitives.asymmetric import rsa as crypto_rsa
from cryptography.hazmat.primitives.ciphers import aead
from cryptography.hazmat.primitives import (
    serialization as crypto_serialization,
)
import jwt

import gcl_iam.constants as c
import gcl_iam.exceptions as exc

LOG = logging.getLogger(__name__)

ALGORITHM_HS256 = c.ALGORITHM_HS256
ALGORITHM_RS256 = c.ALGORITHM_RS256


def _prepare_a256gcm_key(key: tp.Union[str, bytes], key_name: str) -> bytes:
    key_bytes: bytes
    if isinstance(key, bytes):
        key_bytes = key
    else:
        try:
            key_bytes = base64.urlsafe_b64decode(key + "===")
        except (TypeError, binascii.Error):
            key_bytes = key.encode("utf-8")

    if len(key_bytes) != 32:
        raise ValueError(f"{key_name} must be 32 bytes for A256GCM")

    return key_bytes


def encrypt_hs256_jwks_secret(
    secret: str,
    encryption_key: tp.Union[str, bytes],
) -> str:
    key_bytes = _prepare_a256gcm_key(encryption_key, key_name="encryption_key")

    nonce = os.urandom(12)
    ciphertext = aead.AESGCM(key_bytes).encrypt(
        nonce,
        secret.encode("utf-8"),
        None,
    )

    nonce_b64 = base64.urlsafe_b64encode(nonce).rstrip(b"=").decode("ascii")
    ciphertext_b64 = base64.urlsafe_b64encode(ciphertext).rstrip(b"=").decode("ascii")
    return f"aesgcm:{nonce_b64}.{ciphertext_b64}"


def decrypt_hs256_jwks_secret(
    secret: str,
    decryption_key: tp.Union[str, bytes],
) -> str:
    key_bytes = _prepare_a256gcm_key(decryption_key, key_name="decryption_key")

    if not secret.startswith("aesgcm:"):
        raise ValueError("Unsupported encrypted HS256 secret format")

    try:
        nonce_b64, ciphertext_b64 = secret[len("aesgcm:") :].split(".", 1)
    except ValueError:
        raise ValueError("Invalid encrypted HS256 secret format")

    nonce = base64.urlsafe_b64decode(nonce_b64 + "===")
    ciphertext = base64.urlsafe_b64decode(ciphertext_b64 + "===")

    aes = aead.AESGCM(key_bytes)
    plaintext = aes.decrypt(nonce, ciphertext, None)
    return plaintext.decode("utf-8")


def _to_base64url_uint(value: int) -> str:
    raw = value.to_bytes((value.bit_length() + 7) // 8, byteorder="big")
    return base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")


def public_pem_to_jwk(public_key_pem: str) -> dict:
    public_key = crypto_serialization.load_pem_public_key(
        public_key_pem.encode("utf-8")
    )
    if not isinstance(public_key, crypto_rsa.RSAPublicKey):
        raise ValueError("Unsupported public key type")

    public_numbers = public_key.public_numbers()
    jwk: tp.Dict[str, tp.Any] = {
        "kty": "RSA",
        "use": "sig",
        "alg": ALGORITHM_RS256,
        "n": _to_base64url_uint(public_numbers.n),
        "e": _to_base64url_uint(public_numbers.e),
    }
    return jwk


def generate_rsa_private_key_pem(
    bitness: int = 2048,
    public_exponent: int = 65537,
) -> str:
    allowed_bitness = {2048, 3072, 4096}
    if bitness not in allowed_bitness:
        raise ValueError("Invalid RSA bitness. Allowed values: 2048, 3072, 4096.")

    private_key = crypto_rsa.generate_private_key(
        public_exponent=public_exponent,
        key_size=bitness,
    )
    private_key_pem = private_key.private_bytes(
        encoding=crypto_serialization.Encoding.PEM,
        format=crypto_serialization.PrivateFormat.PKCS8,
        encryption_algorithm=crypto_serialization.NoEncryption(),
    )
    return private_key_pem.decode("utf-8")


def _load_rsa_private_key(private_key_pem: str) -> crypto_rsa.RSAPrivateKey:
    private_key = crypto_serialization.load_pem_private_key(
        private_key_pem.encode("utf-8"),
        password=None,
    )
    if not isinstance(private_key, crypto_rsa.RSAPrivateKey):
        raise ValueError("Unsupported private key type")
    return private_key


def generate_rsa_public_key_pem(private_key_pem: str) -> str:
    private_key = _load_rsa_private_key(private_key_pem)

    public_key = private_key.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=crypto_serialization.Encoding.PEM,
        format=crypto_serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return public_key_pem.decode("utf-8")


def get_rsa_bitness_from_private_key_pem(private_key_pem: str) -> int:
    private_key = _load_rsa_private_key(private_key_pem)

    return int(private_key.key_size)


class AbstractAlgorithm(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def decode(self, data: str) -> tp.Dict[str, tp.Any]:
        raise NotImplementedError("Not implemented")

    @abc.abstractmethod
    def encode(self, data: tp.Dict[str, tp.Any]) -> str:
        raise NotImplementedError("Not implemented")


class BaseJwtAlgorithm(AbstractAlgorithm):
    @property
    @abc.abstractmethod
    def algorithm(self) -> str:
        raise NotImplementedError("Not implemented")

    @property
    @abc.abstractmethod
    def candidate_keys(self) -> tp.Iterable[tp.Optional[str]]:
        raise NotImplementedError("Not implemented")

    def _jwt_decode_options(
        self,
        verify: bool,
        ignore_audience: bool,
        ignore_expiration: bool,
    ) -> tp.Dict[str, bool]:
        return {
            "verify_signature": verify,
            "verify_exp": not ignore_expiration,
            "verify_aud": not ignore_audience,
        }

    def _decode_with_fallback_keys(
        self,
        data: str,
        keys: tp.Iterable[tp.Optional[str]],
        algorithm: str,
        options: tp.Dict[str, bool],
        audience: tp.Optional[str],
    ) -> tp.Dict[str, tp.Any]:
        for key in keys:
            if key is None:
                continue
            try:
                return jwt.decode(
                    data,
                    key=key,
                    algorithms=[algorithm],
                    options=options,
                    audience=audience,
                )
            except jwt.exceptions.DecodeError as e:
                LOG.warning("Invalid token by reason: %s", e)
                continue
        raise exc.CredentialsAreInvalidError()

    def decode(
        self,
        data: str,
        audience: tp.Optional[str] = None,
        ignore_audience: bool = False,
        ignore_expiration: bool = False,
        verify: bool = True,
    ) -> tp.Dict[str, tp.Any]:
        options = self._jwt_decode_options(
            verify=verify,
            ignore_audience=ignore_audience,
            ignore_expiration=ignore_expiration,
        )
        return self._decode_with_fallback_keys(
            data,
            keys=self.candidate_keys,
            algorithm=self.algorithm,
            options=options,
            audience=audience,
        )


class HS256(BaseJwtAlgorithm):
    def __init__(self, key: str, previous_key: tp.Optional[str] = None):
        super().__init__()
        self._key = key
        self._previous_key = previous_key

    @property
    def algorithm(self) -> str:
        return ALGORITHM_HS256

    @property
    def candidate_keys(self) -> tp.Iterable[tp.Optional[str]]:
        return (self._key, self._previous_key)

    def encode(self, data: tp.Dict[str, tp.Any]) -> str:
        return jwt.encode(data, key=self._key, algorithm=ALGORITHM_HS256)


class RS256VerifyOnly(BaseJwtAlgorithm):
    def __init__(
        self,
        public_key: str,
        previous_public_key: tp.Optional[str] = None,
    ):
        super().__init__()
        self._public_key = public_key
        self._previous_public_key = previous_public_key

    @property
    def algorithm(self) -> str:
        return ALGORITHM_RS256

    @property
    def candidate_keys(self) -> tp.Iterable[tp.Optional[str]]:
        return (self._public_key, self._previous_public_key)

    def encode(self, data):
        raise NotImplementedError("Signing is not supported by this algorithm")


class RS256(RS256VerifyOnly):
    def __init__(
        self,
        private_key: str,
        public_key: str,
        previous_public_key: tp.Optional[str] = None,
    ):
        super().__init__(
            public_key=public_key,
            previous_public_key=previous_public_key,
        )
        self._private_key = private_key

    def encode(self, data: tp.Dict[str, tp.Any]) -> str:
        return jwt.encode(
            data,
            key=self._private_key,
            algorithm=ALGORITHM_RS256,
        )



#    Copyright 2026 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from restalchemy.api import constants
from restalchemy.api import field_permissions as field_p
from restalchemy.common import contexts

from gcl_iam import rules

# Make it easier to use
Permissions = field_p.Permissions


class FieldsIamPermissions(field_p.BasePermissions):
    def __init__(self, fields, default=Permissions.RW):
        """Field permissions container for resource

        It's based on FieldsPermissions but with Rules support.

        This class describes a dict of fields with permissions for
        various RestAlchemy API methods. All permissions are declared
        in class Permissions. The methods supported by RA are declared
        in the `restalchemy.api.constants` module.
        For example, the code below shows how to set:
        > HIDDEN permissions to the `one_field` for the FILTER method
          and RW for all others RA methods, except FILTER;
        > RO permissions for the `two_field` for all RA methods;
        > HIDDEN permissions for the `three_field` for GET method and
          RO permissions for all others RA methods, except GET.

        In addition, it supports iam permissions check via gcl_iam enforcer.

        Example:

        ```
        FieldsPermissions(
            default=Permissions.RW,
            fields={
                'one_field':{
                    constants.FILTER: Permissions.HIDDEN},
                'two_field': {
                    constants.ALL: Permissions.RO},
                'three_field': {
                    constants.GET: Permissions.HIDDEN,
                    constants.ALL: Permissions.RO},
                'perm_based_field': {
                    constants.ALL: rules.Rule('service', 'res', 'perm')}
            }
        )
        ```

        Pay attention: if you wouldn't set permission for field and RA method
        by default it would be RW (READWRITE) permission.

        :param fields: dict of field model name and permissions
        :param default: default permission for non-described fields
        """
        for method_permission in fields.values():
            for method, permission in method_permission.items():
                assert (
                    method.upper() in constants.ALL_RA_METHODS
                    and permission in Permissions.ALL_PERMISSIONS
                ) or isinstance(permission, rules.Rule)
        super().__init__(permission=default)
        self.fields = fields

    @property
    def _enforcer(self):
        return contexts.get_context().iam_context.enforcer

    def meets_field_permission(self, model_field_name, req, current_permission):

        method = req.api_context.get_active_method()
        field_permission = self.fields.get(model_field_name, {})

        # NOTE(g.melikov): By DEFAULT permission is Permissions.RW
        permission = (
            field_permission.get(method)
            or field_permission.get(constants.ALL)
            or self._permission
        )

        if isinstance(permission, rules.Rule):
            permission = (
                Permissions.RW
                if self._enforcer.enforce(
                    permission,
                    do_raise=False,
                )
                else Permissions.HIDDEN
            )

        return permission <= current_permission



# Copyright 2026 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pytest
from unittest.mock import Mock, patch

from restalchemy.api import constants
from gcl_iam import rules
from gcl_iam.api.field_perms import FieldsIamPermissions, Permissions


def test_init_with_rule_permissions():
    """Test that FieldsIamPermissions accepts rule-based permissions"""
    fields = {
        "field1": {constants.ALL: rules.Rule("service", "resource", "permission")}
    }

    # This should not raise an exception
    permissions = FieldsIamPermissions(fields=fields)

    assert permissions.fields == fields


def test_init_invalid_method():
    """Test that FieldsIamPermissions raises assertion error for invalid methods"""
    fields = {"field1": {"INVALID_METHOD": Permissions.HIDDEN}}

    # This should raise an AssertionError
    with pytest.raises(AssertionError):
        FieldsIamPermissions(fields=fields)


def test_init_invalid_permission():
    """Test that FieldsIamPermissions raises assertion error for invalid permissions"""
    fields = {"field1": {constants.GET: "INVALID_PERMISSION"}}

    # This should raise an AssertionError
    with pytest.raises(AssertionError):
        FieldsIamPermissions(fields=fields)


def test_usual_permissions():
    """Test default permissions without Rules"""
    mock_enforcer = Mock()
    mock_context = Mock()
    mock_context.iam_context.enforcer = mock_enforcer

    fields = {
        "field1": {
            constants.GET: Permissions.HIDDEN,
            constants.CREATE: Permissions.RO,
        }
    }

    mock_req = Mock()
    mock_req.api_context.get_active_method.return_value = constants.CREATE

    with patch("gcl_iam.api.field_perms.contexts.get_context") as mock_get_context:
        mock_get_context.return_value = mock_context
        mock_enforcer.enforce.return_value = True

        permissions = FieldsIamPermissions(fields=fields)

        assert permissions.meets_field_permission("field1", mock_req, Permissions.RW)
        assert permissions.meets_field_permission("field1", mock_req, Permissions.RO)
        assert not permissions.meets_field_permission(
            "field1", mock_req, Permissions.HIDDEN
        )


def test_meets_field_permission_default_permission():
    """Test meets_field_permission with default permission"""
    mock_enforcer = Mock()
    mock_context = Mock()
    mock_context.iam_context.enforcer = mock_enforcer

    fields = {"field1": {constants.CREATE: Permissions.RO}}

    mock_req = Mock()
    mock_req.api_context.get_active_method.return_value = constants.GET

    with patch("gcl_iam.api.field_perms.contexts.get_context") as mock_get_context:
        mock_get_context.return_value = mock_context
        mock_enforcer.enforce.return_value = True

        permissions = FieldsIamPermissions(fields=fields, default=Permissions.HIDDEN)

        # Test with field not in permissions dict - should use default
        result = permissions.meets_field_permission("field2", mock_req, Permissions.RW)
        assert result is True


def test_meets_field_permission_rule_enforcement_positive():
    """Test meets_field_permission with rule enforcement returning positive result"""
    mock_enforcer = Mock()
    mock_context = Mock()
    mock_context.iam_context.enforcer = mock_enforcer

    fields = {
        "field1": {constants.ALL: rules.Rule("service", "resource", "permission")}
    }

    mock_req = Mock()
    mock_req.api_context.get_active_method.return_value = constants.GET

    with patch("gcl_iam.api.field_perms.contexts.get_context") as mock_get_context:
        mock_get_context.return_value = mock_context
        mock_enforcer.enforce.return_value = True

        permissions = FieldsIamPermissions(fields=fields)

        assert permissions.meets_field_permission("field1", mock_req, Permissions.RW)
        assert not permissions.meets_field_permission(
            "field1", mock_req, Permissions.RO
        )
        assert not permissions.meets_field_permission(
            "field1", mock_req, Permissions.HIDDEN
        )


def test_meets_field_permission_rule_enforcement_negative():
    """Test meets_field_permission with rule enforcement returning negative result"""
    mock_enforcer = Mock()
    mock_context = Mock()
    mock_context.iam_context.enforcer = mock_enforcer

    fields = {
        "field1": {constants.ALL: rules.Rule("service", "resource", "permission")}
    }

    mock_req = Mock()
    mock_req.api_context.get_active_method.return_value = constants.GET

    with patch("gcl_iam.api.field_perms.contexts.get_context") as mock_get_context:
        mock_get_context.return_value = mock_context
        mock_enforcer.enforce.return_value = False

        permissions = FieldsIamPermissions(fields=fields)

        assert permissions.meets_field_permission("field1", mock_req, Permissions.RW)
        assert permissions.meets_field_permission("field1", mock_req, Permissions.RO)
        assert permissions.meets_field_permission(
            "field1", mock_req, Permissions.HIDDEN
        )


def test_meets_field_permission_no_permission():
    """Test meets_field_permission when no permissions are set"""
    mock_enforcer = Mock()
    mock_context = Mock()
    mock_context.iam_context.enforcer = mock_enforcer

    fields = {}

    mock_req = Mock()
    mock_req.api_context.get_active_method.return_value = constants.GET

    with patch("gcl_iam.api.field_perms.contexts.get_context") as mock_get_context:
        mock_get_context.return_value = mock_context
        mock_enforcer.enforce.return_value = True

        permissions = FieldsIamPermissions(fields=fields)

        result = permissions.meets_field_permission("field1", mock_req, Permissions.RW)
        assert result is True



#    Copyright 2025 Genesis Corporation.
#    Copyright 2026 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import base64
import dataclasses
import functools
import time
import typing as tp

import bazooka
import bazooka.exceptions
from cryptography.hazmat.primitives.asymmetric import rsa as crypto_rsa
from cryptography.hazmat.primitives import (
    serialization as crypto_serialization,
)
from restalchemy.common import utils

from gcl_iam import algorithms
from gcl_iam import exceptions
from gcl_iam import tokens


class AbstractAuthDriver(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_introspection_info(self, token_info, otp_code=None):
        raise NotImplementedError("Not implemented")

    @abc.abstractmethod
    def get_algorithm(
        self,
        token_info: tokens.UnverifiedToken,
    ) -> algorithms.AbstractAlgorithm:
        raise NotImplementedError("Not implemented")


@dataclasses.dataclass(frozen=True)
class AlgorithmKeys:
    pass


@dataclasses.dataclass(frozen=True)
class HS256AlgorithmKeys(AlgorithmKeys):
    key: str
    previous_key: tp.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class RS256AlgorithmKeys(AlgorithmKeys):
    public_key: str
    previous_public_key: tp.Optional[str] = None


def _base64url_to_int(value: str) -> int:
    decoded = base64.urlsafe_b64decode(value + "===")
    return int.from_bytes(decoded, byteorder="big")


def _rsa_jwk_to_public_key_pem(jwk: tp.Dict[str, tp.Any]) -> str:
    n = jwk["n"]
    e = jwk["e"]

    public_numbers = crypto_rsa.RSAPublicNumbers(
        e=_base64url_to_int(e),
        n=_base64url_to_int(n),
    )
    public_key = public_numbers.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=crypto_serialization.Encoding.PEM,
        format=crypto_serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return public_key_pem.decode("utf-8")


class DummyDriver(AbstractAuthDriver):
    def __init__(self, *args, **kwargs):
        self.reset()

    def reset(self):
        self.user_uuid = "00000000-0000-0000-0000-000000000000"
        self.user_name = "admin"
        self.user_email = "admin@example.com"
        self.user_first_name = "Admin"
        self.user_last_name = "Only For Tests"
        self.project_id = None
        self.otp_verified = True
        self.permission_hash = "00000000-0000-0000-0000-000000000000"
        self.permissions = ["*.*.*"]
        self.algorithm_keys: tp.Dict[str, AlgorithmKeys] = {}

    def get_introspection_info(self, token_info, otp_code=None):
        return {
            "user_info": {
                "uuid": self.user_uuid,
                "name": self.user_name,
                "first_name": self.user_first_name,
                "last_name": self.user_last_name,
                "email": self.user_email,
            },
            "project_id": self.project_id,
            "otp_verified": True if otp_code else self.otp_verified,
            "permission_hash": self.permission_hash,
            "permissions": self.permissions,
        }

    def get_algorithm(
        self,
        token_info: tokens.UnverifiedToken,
    ) -> algorithms.AbstractAlgorithm:
        audience = token_info.audience_name
        if audience not in self.algorithm_keys:
            raise KeyError(f"Unknown audience: {audience}")
        keys = self.algorithm_keys[audience]

        if isinstance(keys, HS256AlgorithmKeys):
            return algorithms.HS256(
                key=keys.key,
                previous_key=keys.previous_key,
            )

        if isinstance(keys, RS256AlgorithmKeys):
            return algorithms.RS256VerifyOnly(
                public_key=keys.public_key,
                previous_public_key=keys.previous_public_key,
            )

        raise TypeError(f"Unexpected algorithm keys type: {type(keys)!r}")


class HttpDriver(AbstractAuthDriver):
    def __init__(
        self,
        iam_endpoint: str,
        audience: str,
        hs256_jwks_decryption_key: str,
        default_timeout=5,
        cache_maxsize: int = 100,
        cache_ttl_seconds: int = 300,
    ):
        super().__init__()
        self._iam_endpoint = utils.lastslash(iam_endpoint)
        self._audience = audience
        self._client = bazooka.Client(default_timeout=default_timeout)
        self._cache_ttl_seconds = cache_ttl_seconds
        self._hs256_jwks_decryption_key = hs256_jwks_decryption_key

        self._get_algorithm_cached = functools.lru_cache(maxsize=cache_maxsize)(
            self._get_algorithm_uncached,
        )

    def get_introspection_info(self, token_info, otp_code=None):
        audience = token_info.audience_name
        if audience != self._audience:
            raise exceptions.TokenAudienceMismatchError(
                token_audience=audience,
                service_audience=self._audience,
            )
        introspection_url = f"{self._iam_endpoint}actions/introspect"
        headers = {"Authorization": f"Bearer {token_info.token}"}
        if otp_code is not None:
            headers["X-OTP"] = otp_code
        try:
            return self._client.get(
                introspection_url,
                headers=headers,
            ).json()
        except bazooka.exceptions.BadRequestError:
            raise exceptions.InvalidAuthTokenError()

    def get_algorithm(
        self,
        token_info: tokens.UnverifiedToken,
    ) -> algorithms.AbstractAlgorithm:
        audience = token_info.audience_name
        if audience != self._audience:
            raise exceptions.TokenAudienceMismatchError(
                token_audience=audience,
                service_audience=self._audience,
            )
        time_bucket = int(time.time() // self._cache_ttl_seconds)
        return self._get_algorithm_cached(time_bucket)

    def _get_algorithm_uncached(
        self,
        time_bucket: int,
    ) -> algorithms.AbstractAlgorithm:
        jwks_url = f"{self._iam_endpoint}actions/jwks"

        payload = self._client.get(
            jwks_url,
        ).json()

        algorithm = payload["algorithm"]
        if algorithm == algorithms.ALGORITHM_HS256:
            hs256_keys = [
                k
                for k in payload["keys"]
                if isinstance(k, dict)
                and k.get("alg") == algorithms.ALGORITHM_HS256
                and k.get("kty") == "oct"
                and k.get("k") is not None
            ]
            if not hs256_keys:
                raise ValueError(
                    "HS256 payload keys list does not contain HS256 oct keys"
                )

            encrypted_key = hs256_keys[0]["k"]
            encrypted_previous_key = None
            if len(hs256_keys) > 1:
                encrypted_previous_key = hs256_keys[1]["k"]

            key = algorithms.decrypt_hs256_jwks_secret(
                secret=encrypted_key,
                decryption_key=self._hs256_jwks_decryption_key,
            )
            previous_key = None
            if encrypted_previous_key is not None:
                previous_key = algorithms.decrypt_hs256_jwks_secret(
                    secret=encrypted_previous_key,
                    decryption_key=self._hs256_jwks_decryption_key,
                )

            return algorithms.HS256(
                key=key,
                previous_key=previous_key,
            )

        elif algorithm == algorithms.ALGORITHM_RS256:
            rs256_keys = [
                k
                for k in payload["keys"]
                if isinstance(k, dict)
                and k.get("alg") == algorithms.ALGORITHM_RS256
                and k.get("kty") == "RSA"
                and k.get("n") is not None
                and k.get("e") is not None
            ]
            if not rs256_keys:
                raise ValueError(
                    "RS256 payload keys list does not contain RS256 RSA keys"
                )

            public_key = _rsa_jwk_to_public_key_pem(rs256_keys[0])
            previous_public_key = None
            if len(rs256_keys) > 1:
                previous_public_key = _rsa_jwk_to_public_key_pem(rs256_keys[1])

            return algorithms.RS256VerifyOnly(
                public_key=public_key,
                previous_public_key=previous_public_key,
            )

        raise ValueError("Unsupported algorithm")



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import os
import unittest.mock as mock

import jwt
import pytest

import gcl_iam.algorithms as algorithms
import gcl_iam.constants as constants
import gcl_iam.drivers as drivers
import gcl_iam.exceptions as exceptions
import gcl_iam.tokens as tokens


def test_rsa_helpers_roundtrip() -> None:
    private_key_pem = algorithms.generate_rsa_private_key_pem(bitness=2048)
    assert isinstance(private_key_pem, str)

    bitness = algorithms.get_rsa_bitness_from_private_key_pem(private_key_pem)
    assert bitness == 2048

    public_key_pem = algorithms.generate_rsa_public_key_pem(private_key_pem)
    assert isinstance(public_key_pem, str)

    jwk = algorithms.public_pem_to_jwk(public_key_pem)
    assert jwk["kty"] == "RSA"
    assert jwk["use"] == "sig"
    assert jwk["alg"] == "RS256"
    assert isinstance(jwk["n"], str)
    assert isinstance(jwk["e"], str)
    assert "=" not in jwk["n"]
    assert "=" not in jwk["e"]


def test_generate_rsa_private_key_pem_invalid_bitness() -> None:
    with pytest.raises(ValueError):
        algorithms.generate_rsa_private_key_pem(bitness=1024)


def test_hs256_jwks_secret_encrypt_decrypt_roundtrip() -> None:
    key_bytes = os.urandom(32)
    key = base64.urlsafe_b64encode(key_bytes).decode("utf-8").rstrip("=")
    encrypted = algorithms.encrypt_hs256_jwks_secret(
        secret="secret",
        encryption_key=key,
    )
    decrypted = algorithms.decrypt_hs256_jwks_secret(
        secret=encrypted,
        decryption_key=key,
    )
    assert decrypted == "secret"


def test_hs256_jwks_secret_encrypt_decrypt_roundtrip_bytes_key() -> None:
    key_bytes = os.urandom(32)
    encrypted = algorithms.encrypt_hs256_jwks_secret(
        secret="secret",
        encryption_key=key_bytes,
    )
    decrypted = algorithms.decrypt_hs256_jwks_secret(
        secret=encrypted,
        decryption_key=key_bytes,
    )
    assert decrypted == "secret"


def test_hs256_encode_decode_current_key() -> None:
    algo = algorithms.HS256(key="current", previous_key="previous")
    payload = {"sub": "user"}

    token = algo.encode(payload)
    decoded = algo.decode(token)

    assert decoded["sub"] == "user"


def test_hs256_decode_previous_key() -> None:
    algo = algorithms.HS256(key="current", previous_key="previous")
    payload = {"sub": "user"}

    token = jwt.encode(payload, key="previous", algorithm=constants.ALGORITHM_HS256)
    decoded = algo.decode(token)

    assert decoded["sub"] == "user"


def test_hs256_invalid_token_raises_credentials_invalid() -> None:
    algo = algorithms.HS256(key="current")

    with pytest.raises(exceptions.CredentialsAreInvalidError):
        algo.decode("not-a-jwt")


def test_rs256_encode_decode_current_key() -> None:
    private_key_pem = algorithms.generate_rsa_private_key_pem(bitness=2048)
    public_key_pem = algorithms.generate_rsa_public_key_pem(private_key_pem)

    algo = algorithms.RS256(private_key=private_key_pem, public_key=public_key_pem)
    payload = {"sub": "user"}

    token = algo.encode(payload)
    decoded = algo.decode(token)

    assert decoded["sub"] == "user"


def test_rs256_decode_previous_public_key() -> None:
    old_private_key_pem = algorithms.generate_rsa_private_key_pem(bitness=2048)
    old_public_key_pem = algorithms.generate_rsa_public_key_pem(old_private_key_pem)

    new_private_key_pem = algorithms.generate_rsa_private_key_pem(bitness=2048)
    new_public_key_pem = algorithms.generate_rsa_public_key_pem(new_private_key_pem)

    token = jwt.encode({"sub": "user"}, key=old_private_key_pem, algorithm="RS256")

    algo = algorithms.RS256(
        private_key=new_private_key_pem,
        public_key=new_public_key_pem,
        previous_public_key=old_public_key_pem,
    )
    decoded = algo.decode(token)

    assert decoded["sub"] == "user"


def test_rs256_invalid_token_raises_credentials_invalid() -> None:
    private_key_pem = algorithms.generate_rsa_private_key_pem(bitness=2048)
    public_key_pem = algorithms.generate_rsa_public_key_pem(private_key_pem)

    algo = algorithms.RS256(private_key=private_key_pem, public_key=public_key_pem)

    with pytest.raises(exceptions.CredentialsAreInvalidError):
        algo.decode("not-a-jwt")


def test_rs256_decode_calls_jwt_decode() -> None:
    private_key_pem = algorithms.generate_rsa_private_key_pem(bitness=2048)
    public_key_pem = algorithms.generate_rsa_public_key_pem(private_key_pem)

    algo = algorithms.RS256(private_key=private_key_pem, public_key=public_key_pem)

    with mock.patch.object(jwt, "decode", wraps=jwt.decode) as jwt_decode:
        token = algo.encode({"sub": "user"})
        algo.decode(token)

    assert jwt_decode.called


def test_rs256_verify_only_decode_ok() -> None:
    private_key_pem = algorithms.generate_rsa_private_key_pem(bitness=2048)
    public_key_pem = algorithms.generate_rsa_public_key_pem(private_key_pem)

    token = jwt.encode({"sub": "user"}, key=private_key_pem, algorithm="RS256")
    algo = algorithms.RS256VerifyOnly(public_key=public_key_pem)

    decoded = algo.decode(token)

    assert decoded["sub"] == "user"


def test_rs256_verify_only_encode_raises() -> None:
    private_key_pem = algorithms.generate_rsa_private_key_pem(bitness=2048)
    public_key_pem = algorithms.generate_rsa_public_key_pem(private_key_pem)

    algo = algorithms.RS256VerifyOnly(public_key=public_key_pem)

    with pytest.raises(NotImplementedError):
        algo.encode({"sub": "user"})


def test_http_driver_get_algorithm_keys_hs256_jwks_payload_ok() -> None:
    aes_key = os.urandom(32)
    aes_key_b64 = base64.urlsafe_b64encode(aes_key).decode("utf-8").rstrip("=")
    encrypted_secret = algorithms.encrypt_hs256_jwks_secret(
        secret="secret",
        encryption_key=aes_key_b64,
    )

    driver = drivers.HttpDriver(
        "http://iam.example/",
        audience="client-1",
        hs256_jwks_decryption_key=aes_key_b64,
    )
    driver._client = mock.Mock()
    driver._client.get.return_value.json.return_value = {
        "keys": [
            {
                "kty": "oct",
                "alg": "HS256",
                "use": "sig",
                "kid": "00000000-0000-0000-0000-000000000001",
                "k": encrypted_secret,
            },
        ],
        "algorithm": algorithms.ALGORITHM_HS256,
    }

    token_info = mock.Mock(spec=tokens.UnverifiedToken)
    token_info.audience_name = "client-1"

    algo = driver.get_algorithm(token_info)
    assert isinstance(algo, algorithms.HS256)

    token = algo.encode({"sub": "user"})
    decoded = algo.decode(token)
    assert decoded["sub"] == "user"


def test_http_driver_get_algorithm_rs256_verify_only_jwks_payload_ok() -> None:
    aes_key = os.urandom(32)
    aes_key_b64 = base64.urlsafe_b64encode(aes_key).decode("utf-8").rstrip("=")

    private_key_pem = algorithms.generate_rsa_private_key_pem(bitness=2048)
    public_key_pem = algorithms.generate_rsa_public_key_pem(private_key_pem)
    jwk = algorithms.public_pem_to_jwk(public_key_pem)

    driver = drivers.HttpDriver(
        "http://iam.example/",
        audience="client-1",
        hs256_jwks_decryption_key=aes_key_b64,
    )
    driver._client = mock.Mock()
    driver._client.get.return_value.json.return_value = {
        "algorithm": algorithms.ALGORITHM_RS256,
        "keys": [jwk],
    }

    token_info = mock.Mock(spec=tokens.UnverifiedToken)
    token_info.audience_name = "client-1"

    algo = driver.get_algorithm(token_info)
    assert isinstance(algo, algorithms.RS256VerifyOnly)

    token = jwt.encode({"sub": "user"}, key=private_key_pem, algorithm="RS256")
    decoded = algo.decode(token)
    assert decoded["sub"] == "user"


def test_http_driver_get_algorithm_audience_mismatch_raises() -> None:
    aes_key = os.urandom(32)
    aes_key_b64 = base64.urlsafe_b64encode(aes_key).decode("utf-8").rstrip("=")

    driver = drivers.HttpDriver(
        "http://iam.example/",
        audience="client-1",
        hs256_jwks_decryption_key=aes_key_b64,
    )
    driver._client = mock.Mock()

    token_info = mock.Mock(spec=tokens.UnverifiedToken)
    token_info.audience_name = "client-2"

    with pytest.raises(exceptions.TokenAudienceMismatchError):
        driver.get_algorithm(token_info)

    assert not driver._client.get.called


def test_http_driver_get_introspection_info_audience_mismatch_raises() -> None:
    aes_key = os.urandom(32)
    aes_key_b64 = base64.urlsafe_b64encode(aes_key).decode("utf-8").rstrip("=")

    driver = drivers.HttpDriver(
        "http://iam.example/",
        audience="client-1",
        hs256_jwks_decryption_key=aes_key_b64,
    )
    driver._client = mock.Mock()

    token_info = mock.Mock(spec=tokens.UnverifiedToken)
    token_info.audience_name = "client-2"
    token_info.token = "dummy"

    with pytest.raises(exceptions.TokenAudienceMismatchError):
        driver.get_introspection_info(token_info)

    assert not driver._client.get.called



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import logging
import urllib.parse

from restalchemy.common import contexts
from restalchemy.storage.sql import engines

from gcl_iam import constants as c
from gcl_iam import exceptions as e

LOG = logging.getLogger(__name__)


class GenesisCoreAuthContext(contexts.ContextWithStorage):
    def __init__(
        self,
        req,
        engine_name: str = engines.DEFAULT_NAME,
        context_storage: contexts.Storage = None,
    ):
        super().__init__(engine_name, context_storage)
        self._req = req

    @property
    def request(self):
        return self._req

    def get_real_url_with_prefix(self):
        headers = self._req.headers
        fallback_url = self._req.host_url

        forwarded_proto = headers.get("X-Forwarded-Proto")
        forwarded_host = headers.get("X-Forwarded-Host")
        forwarded_port = headers.get("X-Forwarded-Port")
        forwarded_prefix = headers.get("X-Forwarded-Prefix")

        parsed = urllib.parse.urlsplit(fallback_url)

        scheme = forwarded_proto or parsed.scheme
        host = forwarded_host or parsed.hostname
        port = forwarded_port or parsed.port

        new_uri = (
            f"{scheme}://{host}:{port}" if port is not None else f"{scheme}://{host}"
        )

        if forwarded_prefix:
            new_uri += forwarded_prefix.rstrip("/")

        return new_uri

    @contextlib.contextmanager
    def iam_session(self, iam_context):
        self._store_iam_session(iam_context)
        try:
            LOG.debug("Start iam session with context: %s", iam_context)
            yield iam_context
        finally:
            LOG.debug("End iam session with context: %s", iam_context)
            self._remove_iam_session()

    @property
    def iam_context(self):
        self._check_iam_session()
        return getattr(self._local_thread_storage, c.CONTEXT_STORAGE_KEY)

    def _store_iam_session(self, iam_contex):
        if hasattr(self._local_thread_storage, c.CONTEXT_STORAGE_KEY):
            raise e.AnotherIamSessionAlreadyStored()

        setattr(self._local_thread_storage, c.CONTEXT_STORAGE_KEY, iam_contex)

    def _check_iam_session(self):
        if not hasattr(self._local_thread_storage, c.CONTEXT_STORAGE_KEY):
            raise e.NoIamSessionStored()

    def _remove_iam_session(self):
        self._check_iam_session()
        delattr(self._local_thread_storage, c.CONTEXT_STORAGE_KEY)
